package ch.ethz.icr.tutorial.evolipd;

import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.SimpleModel;
import uchicago.src.sim.util.SimUtilities;

/**
 * Evolutionary game based on "soup"
 *
 * This is the second step of the tutorial. This stage in the tutorial sequence
 * adds dynamics to the four strategies. The model features an arbitrary number
 * of players that encounter each other randomly. An adaptation mechanism selects
 * out the most successful players.
 *
 * For the original model, see Cohen, Riolo, and Axelrod:
 * The Emergence of Social Organization in the Prisoner's Dilemma
 * (SFI Working Paper), 1999. http://www.santafe.edu/ (follow publications link).
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @author Laszlo Gulyas
 * @version 2.0
 */
public final class Model extends SimpleModel {

    // Constants for agent strategies
    private static final int ALLC = 0;
    private static final int TFT = 1;
    private static final int ATFT = 2;
    private static final int ALLD = 3;

    // Number of possible strategies (used when looping through all possible
    // strategies).
    private static final int NUMTYPES = 4;

    // Model variables
    private int numPlayers;             // Total number of agents
    private double pALLC;
    private double pTFT;
    private double pATFT;
    private double pALLD;               // Fractions for each strategy type
    private int[] num;                  // Array holding number of agents for
                                        // each strategy type

    private int maxIter;                // Number of iteration cycles
    private int numNeighbors;           // The number of opponents to be met in
                                        // each game
    double pAdapt;                      // Probability of adaptation
                                        // Note that this feature is an
                                        // extension of the original Cohen et al.
                                        // model. pAdapt=1.0 yields the original
                                        // behavior.

    /**
     * Initializing the model.
     */
    public final void setup() {
        // The setup() method of the parent MUST be called first!
        super.setup();

        // Setting the name of our model
        name = "EvolIPD Model (Tutorial Step #2)";

        // Setting the values of the model parameters
        numPlayers = 256;
        pALLC = 0.25;
        pTFT = 0.25;
        pATFT = 0.25;
        pALLD = 0.25;

        maxIter = 4;
        numNeighbors = 4;
        pAdapt = 0.2;
    }

    /**
     * The method to build the Model's internals.
     */
    public final void buildModel() {
        // Determining the number of agents in each types
        // Create the array holding the number of players per type
        num = new int[4];
        // Calculate these numbers
        num[ALLC] = (int) (numPlayers * pALLC);
        num[TFT] = (int) (numPlayers * pTFT);
        num[ATFT] = (int) (numPlayers * pATFT);
        num[ALLD] = (int) (numPlayers * pALLD);

        // Making sure we avoid cases when truncation leads to less than
        // numPlayers agents.
        if (num[ALLD] != numPlayers - num[ALLC] - num[TFT] - num[ATFT])
            num[ALLD] = numPlayers - num[ALLC] - num[TFT] - num[ATFT];

        // Creating the agents
        int playerID = 0;   // The ID of the agent under creation

        // We double-loop through each strategy and number of players
        // per strategy.
        // First, loop through all the different types
        for (int playerType = ALLC; playerType < NUMTYPES; playerType++)
                // Creating the appropriate number of agents
            for (int i = 0; i < num[playerType]; i++) {
                // Increment the identity counter
                playerID++;

                // Create a player
                final Player aPlayer = new Player(playerID, playerType, this);

                // Add the newly created agent to the list of agents
                agentList.add(aPlayer);
            }
        // Use Repast's shuffle method to randomize the order of the agents
        SimUtilities.shuffle(agentList);

        // Show the initial statistics
        reportResults();
    }

    /////////////////////////////////////////////////////////////////////////
    //  Iterated methods  ///////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////

    /**
     * The main activity of the time step.
     */
    public final void step() {
        // We carry out four sub-activities:
        resetPlayers();     // Reset the agents' statistics
        interactions();     // Let them interact with each other
        adaptation();       // Let them adapt
        reportResults();    // Calculate and report some statistics
    }

    /**
     * Sub-activity: Resetting the agents.
     */
    private void resetPlayers() {
        // Loop through the entire agent list
        for (int i = 0; i < numPlayers; i++) {
            // Pick an agent
            final Player aPlayer = (Player) agentList.get(i);

            // Let it reset it's stats
            aPlayer.reset();
        }
    }

    /**
     * Sub-activity: Interactions.
     * The interaction method lets each player play against as many as
     * numNeighbors others
     */
    private void interactions() {
        // Looping through all the agents
        for (int i = 0; i < numPlayers; i++) {
            final Player aPlayer = (Player) agentList.get(i);  // Get one player

            // Looping through the appropriate number of neighbors
            for (int j = 0; j < numNeighbors; j++) {
                int randomIndex;
                Player other;

                // Randomly selecting an agent, so that it is
                // different from the current one (aPlayer)
                // LOOP BEGIN: Keep trying...
                do {
                    // Draw a uniform random integer (method of SimpleModel)
                    randomIndex = getNextIntFromTo(0, numPlayers - 1);

                    // Get the player
                    other = (Player) agentList.get(randomIndex);
                } while (other.playerID == aPlayer.playerID);
                // LOOP END: until a different one is found

                // Play the two-person game between the two agents
                game(aPlayer, other);
            }
        }
    }

    /**
     * A two-person game.
     *
     */
    private void game(final Player rowPlayer, final Player colPlayer) {
        // First we need to tell the two players about each other's existence and
        // add the other player to their neighborhood memory (i.e. otherList).
        // The latter will be used in the adaptation method.
        rowPlayer.other = colPlayer;
        rowPlayer.otherList.add(colPlayer);
        colPlayer.other = rowPlayer;
        colPlayer.otherList.add(rowPlayer);

        // Here the iterated game unfolds in maxIter rounds as in SimpleIPD
        for (int i = 1; i <= maxIter; i++) {    // Note that 'time' starts at 1.
            rowPlayer.play(i);
            colPlayer.play(i);
            rowPlayer.remember();
            colPlayer.remember();
            rowPlayer.addPayoff();
            colPlayer.addPayoff();
        }
    }

    /**
     * Sub-activity: Let the agents adapt themselves to the best strategy
     * in their neighborhood (what they have recorded).
     */
    private void adaptation() {
        // NOTE DOUBLE BUFFERING!
        // Let all agents calculate their adapted type first
        for (int i = 0; i < numPlayers; i++) {
            final Player aPlayer = (Player) agentList.get(i);
            aPlayer.adapt();
        }
        // Second, once they have all calculated their new strategy,
        // let them update to the new type
        for (int i = 0; i < numPlayers; i++) {
            final Player aPlayer = (Player) agentList.get(i);
            aPlayer.updateType();
        }
    }

    /**
     * Sub-activity: Calculating and outputting the statistics.
     */
    private void reportResults() {
        // Initializing the temporary variables (the array of counters)
        for (int i = ALLC; i < NUMTYPES; i++)
            num[i] = 0;

        // Counting the number of players per strategy type by incrementing
        // the appropriate counter
        for (int i = 0; i < numPlayers; i++) {
            final Player aPlayer = (Player) agentList.get(i);
            num[aPlayer.type]++;
        }

        // Printing the results to the screen
        System.out.print(getTickCount() + ": ");
        for (int playerType = ALLC; playerType < NUMTYPES; playerType++)
            System.out.print(num[playerType] + " ");
        System.out.println();
    }

    /**
     * Creating and starting your model.
     * 
     */
    public static void main(final String[] args) {
        final SimInit init = new SimInit();
        final Model m = new Model();
        init.loadModel(m, null, false);
    }
}
