package ch.ethz.icr.tutorial.evolipd;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Evolutionary game based on "soup": The agent's code
 * <p/>
 * This is the second step of the tutorial. This stage in the tutorial sequence
 * adds dynamics to the four strategies. The model features an arbitrary number
 * of players that encounter each other randomly. An adaptation mechanism selects
 * out the most successful players.
 * <p/>
 * For the original model, see Cohen, Riolo, and Axelrod:
 * The Emergence of Social Organization in the Prisoner's Dilemma
 * (SFI Working Paper), 1999. http://www.santafe.edu/ (follow publications link).
 *
 * @author Luc Girardin
 * @author Lars-Erik Cederman
 * @author Laszlo Gulyas
 * @version 2.0
 */
final class Player {
    // Constants for actions
    private static final int C = 1;          // Cooperate
    private static final int D = 0;          // Defect

    // Action matrices
    // The x_PARAMS[] matrices define the behavior for each of the four strategies
    // (i.e., ALLC, TFT, ATFT, ALLD, respectively).
    private final int[] I_PARAMS = {C, C, D, D}; // First action
    private final int[] P_PARAMS = {C, C, D, D}; // Action if opponent cooperated
    private final int[] Q_PARAMS = {C, D, C, D}; // Action if opponent defected

    // The agent's internal variables
    private final Model model;               // Reference (handle) to the model
    final int playerID;                      // The player's identification
    Player other;                            // Handle to the opponent
    int type;                                // The agent's strategy
    private int newType;                     // The agent's calculated strategy
    // during adaptation
    private final int[][] prefs = {{1, 5},   // Payoff / Preference matrix
                     {0, 3}};                // (this one is PD)
    private int action;                      // The current action
    private int memory;                      // The opponent's last action
    private int cumulPayoff;                 // The agent's cumulated payoff
    private int numPlays;                    // Number of games played
                                             // (for statistics)

    final ArrayList otherList;               // List of opponents

    /**
     * Creating the agent.
     *
     */
    public Player(final int i, final int t, final Model m) {
        playerID = i;
        type = t;
        model = m;
        otherList = new ArrayList();
    }

    /**
     * Initializing the agent's variables (all counters have to be reset).
     */
    public final void reset() {
        numPlays = 0;
        cumulPayoff = 0;
        otherList.clear();
    }

    /**
     * Storing the opponent's last action (same as that in SimpleIPD).
     */
    public final void remember() {
        memory = other.action;
    }

    /**
     * The agent's decision-making.
     * This is the main behavioral method that uses one of the x_PARAMS[][]
     * matrices depending on whether the move is the first one or the other
     * player cooperated. The actual move is also dependent on the agent's
     * type. The move is recorded in the player's action variable.
     * <p/>
     * The play method is identical to that in SimpleIPD except that the numPlay
     * counter has to be incremented.
     *
     */
    public final void play(final int time) {
        // Keeping track of the number of games played
        numPlays++;
        if (time == 1)
            action = I_PARAMS[type];
        else if (memory == C)
            action = P_PARAMS[type];
        else
            action = Q_PARAMS[type];
    }

    /**
     * Administering the payoff.
     * Updates the cumulative payoff by adding the payoffs from the preference
     * matrix as a function of both sides' moves (same as that in SimpleIPD).
     */
    public final void addPayoff() {
        cumulPayoff = cumulPayoff + prefs[action][other.action];
    }

    // In this pair of methods the player updates the strategy to that of the
    // most successful player encountered (based on the otherList).

    /**
     * Calculating the agent's new strategy.
     */
    public final void adapt() {
        // We use double-buffering by storing the result in newType which will
        // later be used to update type (but we can't to do that until we've gone
        // through all players):
        newType = type;

        // We use SimpleModel's uniform random number generator to draw a random
        // number and with probability pAdapt we let the player execute the body
        // of the method
        if (model.getNextDoubleFromTo(0.0, 1.0) < model.pAdapt) {
            // This is a simple search loop going through the otherList to find
            // the best playoff among the opponents. The first candidate for the
            // best payoff player is the agent itself.
            double bestPayoff = getAveragePayoff();
            final Iterator i = otherList.iterator();
            while (i.hasNext()) {
                final Player aPlayer = (Player) i.next();
                final double payoff = aPlayer.getAveragePayoff();
                if (payoff > bestPayoff) {
                    bestPayoff = payoff;
                    // Set the new type to the best known (up to now)
                    newType = aPlayer.type;
                }
            }
        }
    }

    /**
     * Complete double-buffering by updating the strategy type to the
     * recently calculated value.
     */
    public final void updateType() {
        type = newType;
    }

    /**
     * Helper function returning the agent's average cumulative payoff.
     *
     */
    private double getAveragePayoff() {
        if (numPlays == 0)          // Check that avoids division by zero
            return -1.0;              // Extreme value as an 'error message'
        else
            return (double) cumulPayoff / (double) numPlays;
    }
}